@@ -36,86 +36,28 @@ import sys
 import time
 import hashlib
 
-if int(sys.version[0]) == 3:
+if sys.version_info[0] >= 3:
 	import urllib.request, urllib.parse, urllib.error
-	url_parse = urllib.parse
+	url_parse = urllib.parse.urlparse
+	url_unparse = urllib.parse.urlunparse
 	url_open = urllib.request.urlopen
+	HTTPError = urllib.error.HTTPError
+	import http.client
+	IncompleteRead = http.client.IncompleteRead
 else:
-	import urllib
+	import urllib2
 	import urlparse
 	url_parse = urlparse.urlparse
-	url_open = urllib.urlopen
+	url_unparse = urlparse.urlunparse
+	url_open = urllib2.urlopen
+	HTTPError = urllib2.HTTPError
+	import httplib
+	IncompleteRead = httplib.IncompleteRead
 
 
-from mirrorselect.mirrorparser3 import MirrorParser3
 from mirrorselect.output import encoder, get_encoding, decode_selection
 
 
-class Extractor(object):
-	"""The Extractor employs a MirrorParser3 object to get a list of valid
-	mirrors, and then filters them. Only the mirrors that should be tested, based on
-	user input are saved. They will be in the hosts attribute."""
-
-	def __init__(self, list_url, options, output):
-		self.output = output
-		parser = MirrorParser3()
-		self.hosts = []
-
-		hosts = self.getlist(parser, list_url)
-		self.output.write('Extractor(): fetched mirrors.xml,'
-				' %s hosts before filtering\n' % len(hosts), 2)
-
-		if not options.rsync:
-			if options.ftp:
-				hosts = self.restrict_protocall('ftp', hosts)
-			if options.http:
-				hosts = self.restrict_protocall('http', hosts)
-
-		self.hosts = hosts
-
-	def restrict_protocall(self, prot, hosts):
-		"""
-		Removes hosts that are not of the specified type.
-		"prot" must always be exactly 'http' or 'ftp'.
-		"""
-		myhosts = []
-
-		self.output.print_info('Limiting test to %s hosts. ' % prot )
-
-		for host in hosts:
-			if host[0].startswith(prot):
-				myhosts.append(host)
-
-		self.output.write('%s of %s removed.\n' % (len(hosts) - len(myhosts),
-			len(hosts)) )
-
-		return myhosts
-
-
-	def getlist(self, parser, url):
-		"""
-		Uses the supplied parser to get a list of urls.
-		Takes a parser object, url, and filering options.
-		"""
-
-		self.output.write('getlist(): fetching ' + url + '\n', 2)
-
-		self.output.print_info('Downloading a list of mirrors...')
-
-		try:
-			parser.parse(url_open(url).read())
-		except EnvironmentError:
-			pass
-
-		if len(parser.tuples()) == 0:
-			self.output.print_err('Could not get mirror list. Check your internet'
-					' connection.')
-
-		self.output.write(' Got %d mirrors.\n' % len(parser.tuples()))
-
-		return parser.tuples()
-
-
 class Shallow(object):
 	"""handles rapid server selection via netselect"""
 
@@ -144,13 +86,13 @@ class Shallow(object):
 		top_hosts = []
 
 		if not quiet:
-			self.output.print_info('Using netselect to choose the top %d mirrors...' \
-					% number)
+			self.output.print_info('Using netselect to choose the top '
+				'%d mirrors...' % number)
 
 		host_string = ' '.join(hosts)
 
-		self.output.write('\nnetselect(): running "netselect -s%d %s"\n' % (int(number),
-			host_string), 2)
+		self.output.write('\nnetselect(): running "netselect -s%d %s"\n'
+			% (int(number), host_string), 2)
 
 		proc = subprocess.Popen( ['netselect', '-s%d' % (number,)] + hosts,
 			stdout=subprocess.PIPE, stderr=subprocess.PIPE)
@@ -181,7 +123,8 @@ class Shallow(object):
 
 	def netselect_split(self, hosts, number, block_size):
 		"""
-		This uses netselect to test mirrors in chunks, each at most block_size in length.
+		This uses netselect to test mirrors in chunks,
+		each at most block_size in length.
 		This is done in a tournament style.
 		"""
 		hosts = [host[0] for host in hosts]
@@ -203,8 +146,8 @@ class Shallow(object):
 
 			host_dict = self.netselect(block, len(block), quiet=True)[1]
 
-			self.output.write('ran netselect(%s, %s), and got %s\n' % (block, len(block),
-				host_dict), 2)
+			self.output.write('ran netselect(%s, %s), and got %s\n'
+				% (block, len(block), host_dict), 2)
 
 			for key in list(host_dict.keys()):
 				ret_hosts[key] = host_dict[key]
@@ -227,7 +170,8 @@ class Shallow(object):
 
 	def host_blocks(self, hosts, block_size):
 		"""
-		Takes a list of hosts and a block size, and returns an list of lists of URLs.
+		Takes a list of hosts and a block size,
+		and returns an list of lists of URLs.
 		Each of the sublists is at most block_size in length.
 		"""
 		host_array = []
@@ -240,12 +184,21 @@ class Shallow(object):
 			mylist = []
 		host_array.append(hosts)
 
-		self.output.write('\n_host_blocks(): returns %s blocks, each about %s in size\n'
-				% (len(host_array), len(host_array[0])), 2)
+		self.output.write('\n_host_blocks(): returns '
+			'%s blocks, each about %s in size\n'
+			% (len(host_array), len(host_array[0])), 2)
 
 		return host_array
 
 
+class TimeoutException(Exception):
+	pass
+
+
+def timeout_handler(signum, frame):
+	raise TimeoutException()
+
+
 class Deep(object):
 	"""handles deep mode mirror selection."""
 
@@ -257,6 +210,8 @@ class Deep(object):
 		self._dns_timeout = options.timeout
 		self._connect_timeout = options.timeout
 		self._download_timeout = options.timeout
+		self.test_file = options.file
+		self.test_md5 = options.md5
 
 		addr_families = []
 		if options.ipv4:
@@ -264,9 +219,7 @@ class Deep(object):
 		elif options.ipv6:
 			addr_families.append(socket.AF_INET6)
 		else:
-			addr_families.append(socket.AF_INET)
-			if socket.has_ipv6:
-				addr_families.append(socket.AF_INET6)
+			addr_families.append(socket.AF_UNSPEC)
 
 		self._addr_families = addr_families
 
@@ -283,12 +236,19 @@ class Deep(object):
 		maxtime = self._download_timeout
 		hosts = [host[0] for host in self._hosts]
 		num_hosts = len(hosts)
+		self.dl_failures = 0
 
 		for host in hosts:
 
 			prog += 1
-			self.output.print_info('Downloading 100k files from each mirror... [%s of %s]'\
-							% (prog, num_hosts) )
+			if self.test_file is not 'mirrorselect-test':
+				self.output.print_info(
+					'Downloading %s files from each mirror... [%s of %s]'
+					% (self.test_file, prog, num_hosts) )
+			else:
+				self.output.print_info(
+					'Downloading 100k files from each mirror... [%s of %s]'
+					% (prog, num_hosts) )
 
 			mytime, ignore = self.deeptime(host, maxtime)
 
@@ -298,21 +258,25 @@ class Deep(object):
 			else:
 				continue
 
-		self.output.write('deeptest(): got %s hosts, and returned %s\n' % (num_hosts, \
-			str(list(top_hosts.values()))), 2)
+		self.output.write('deeptest(): got %s hosts, and returned %s\n'
+			% (num_hosts, str(list(top_hosts.values()))), 2)
 
 		self.output.write('\n')	#this just makes output nicer
 
-		#can't just return the dict.values, because we want the fastest mirror first...
+		#can't just return the dict.values,
+		#because we want the fastest mirror first...
 		keys = list(top_hosts.keys())
 		keys.sort()
 
 		rethosts = []
 		for key in keys:
-			#self.output.write('deeptest(): adding rethost %s, %s' % (key, top_hosts[key]), 2)
+			#self.output.write('deeptest(): adding rethost '
+				#'%s, %s' % (key, top_hosts[key]), 2)
 			rethosts.append(top_hosts[key])
 
-		self.output.write('deeptest(): final rethost %s' % (rethosts), 2)
+		self.output.write('deeptest(): final rethost %s\n' % (rethosts), 2)
+		self.output.write('deeptest(): final md5 failures %s of %s\n'
+			% (self.dl_failures, num_hosts), 2)
 		self.urls = rethosts
 
 
@@ -325,66 +289,53 @@ class Deep(object):
 		self.output.write('\n_deeptime(): maxtime is %s\n' % maxtime, 2)
 
 		if url.endswith('/'):	#append the path to the testfile to the URL
-			url = url + 'distfiles/mirrorselect-test'
+			url = url + 'distfiles/' + self.test_file
 		else:
-			url = url + '/distfiles/mirrorselect-test'
+			url = url + '/distfiles/' + self.test_file
 
 		url_parts = url_parse(url)
 
-		class TimeoutException(Exception):
-			pass
-
-		def timeout_handler(signum, frame):
-			raise TimeoutException()
-
 		signal.signal(signal.SIGALRM, timeout_handler)
 
 		ips = []
-		for family in self._addr_families:
-			ipv6 = family == socket.AF_INET6
+		for addr_family in self._addr_families:
 			try:
-				try:
-					signal.alarm(self._dns_timeout)
-					for family, socktype, proto, canonname, sockaddr in \
-						socket.getaddrinfo(url_parts.hostname, None,
-							family, socket.SOCK_STREAM):
-						ip = sockaddr[0]
-						if ipv6:
-							ip = "[%s]" % ip
-						ips.append(ip)
-				finally:
-					signal.alarm(0)
+				signal.alarm(self._dns_timeout)
+				for result in socket.getaddrinfo(url_parts.hostname, None,
+					addr_family, socket.SOCK_STREAM, 0, socket.AI_ADDRCONFIG):
+					family, _, __, ___, sockaddr = result
+					ip = sockaddr[0]
+					if family == socket.AF_INET6:
+						ip = "[%s]" % ip
+					ips.append(ip)
 			except socket.error as e:
-				self.output.write('deeptime(): dns error for host %s: %s\n' % \
-					(url_parts.hostname, e), 2)
+				self.output.write('deeptime(): dns error for host %s: %s\n'
+					% (url_parts.hostname, e), 2)
 			except TimeoutException:
-				self.output.write('deeptime(): dns timeout for host %s\n' % \
-					(url_parts.hostname,), 2)
+				self.output.write('deeptime(): dns timeout for host %s\n'
+					% url_parts.hostname, 2)
+			finally:
+				signal.alarm(0)
 
 		if not ips:
-			self.output.write('deeptime(): unable to resolve ip for host %s\n' % \
-				(url_parts.hostname,), 2)
+			self.output.write('deeptime(): unable to resolve ip for host %s\n'
+				% url_parts.hostname, 2)
 			return (None, True)
 
+		self.output.write("deeptime(): ip's for host %s: %s\n"
+			% (url_parts.hostname, str(ips)), 2)
 		delta = 0
 		f = None
 
 		for ip in ips:
-			try:
-				try:
-					signal.alarm(self._connect_timeout)
-					f = url_open(url)
-					break
-				finally:
-					signal.alarm(0)
-			except EnvironmentError as e:
-				self.output.write(('deeptime(): connection to host %s ' + \
-					'failed for ip %s: %s\n') % \
-					(url_parts.hostname, ip, e), 2)
-			except TimeoutException:
-				self.output.write(('deeptime(): connection to host %s ' + \
-					'timed out for ip %s\n') % \
-					(url_parts.hostname, ip), 2)
+			test_parts = url_parts._replace(netloc=ip)
+			test_url = url_unparse(test_parts)
+			self.output.write('deeptime(): testing url: %s\n' % test_url, 2)
+
+			f, test_url, early_out = self._test_connection(test_url, url_parts,
+				ip, ips[ips.index(ip):])
+			if early_out:
+				break
 
 		if f is None:
 			self.output.write('deeptime(): unable to ' + \
@@ -400,14 +351,13 @@ class Deep(object):
 			finally:
 				signal.alarm(0)
 		except EnvironmentError as e:
-			self.output.write(('deeptime(): close connection to host %s ' + \
-				'failed for ip %s: %s\n') % \
-				(url_parts.hostname, ip, e), 2)
+			self.output.write(('deeptime(): closing connection to host %s '
+				'failed for ip %s: %s\n') % (url_parts.hostname, ip, e), 2)
 		except TimeoutException:
-			self.output.write(('deeptime(): close connection to host %s ' + \
-				'timed out for ip %s\n') % \
-				(url_parts.hostname, ip), 2)
+			self.output.write(('deeptime(): closing connection to host %s '
+				'timed out for ip %s\n') % (url_parts.hostname, ip), 2)
 
+		self.output.write('deeptime(): timing url: %s\n' % test_url, 2)
 		try:
 			# The first connection serves to "wake up" the route between
 			# the local and remote machines. A second connection is used
@@ -415,56 +365,103 @@ class Deep(object):
 			try:
 				signal.alarm(int(math.ceil(maxtime)))
 				stime = time.time()
-				f = url_open(url)
+				f = url_open(test_url)
 
-				if hashlib.md5(f.read()).hexdigest() != "bdf077b2e683c506bf9e8f2494eeb044":
-					return (None, True)
+				md5 = hashlib.md5(f.read()).hexdigest()
 
 				delta = time.time() - stime
 				f.close()
+				if md5 != self.test_md5:
+					self.output.write(
+						"\ndeeptime(): md5sum error for file: %s\n"
+						% self.test_file +
+						"         expected: %s\n" % self.test_md5 +
+						"         got.....: %s\n" % md5 +
+						"         host....: %s, %s\n"
+						% (url_parts.hostname, ip))
+					self.dl_failures += 1
+					return (None, True)
+
 			finally:
 				signal.alarm(0)
 
 		except EnvironmentError as e:
-			self.output.write(('deeptime(): download from host %s ' + \
-				'failed for ip %s: %s\n') % \
-				(url_parts.hostname, ip, e), 2)
+			self.output.write(('\ndeeptime(): download from host %s '
+				'failed for ip %s: %s\n') % (url_parts.hostname, ip, e), 2)
 			return (None, True)
 		except TimeoutException:
-			self.output.write(('deeptime(): download from host %s ' + \
-				'timed out for ip %s\n') % \
-				(url_parts.hostname, ip), 2)
+			self.output.write(('\ndeeptime(): download from host %s '
+				'timed out for ip %s\n') % (url_parts.hostname, ip), 2)
+			return (None, True)
+		except IncompleteRead as e:
+			self.output.write(('\ndeeptime(): download from host %s '
+				'failed for ip %s: %s\n') % (url_parts.hostname, ip, e), 2)
 			return (None, True)
 
 		signal.signal(signal.SIGALRM, signal.SIG_DFL)
 
 		self.output.write('deeptime(): download completed.\n', 2)
-		self.output.write('deeptime(): %s seconds for host %s\n' % (delta, url), 2)
+		self.output.write('deeptime(): %s seconds for host %s\n'
+			% (delta, url), 2)
 		return (delta, False)
 
+
+	def _test_connection(self, test_url, url_parts, ip, ips):
+		"""Tests the url for a connection, will recurse using
+		the original url instead of the ip if an HTTPError occurs
+		Returns f, test_url, early_out
+		"""
+		early_out = False
+		f = None
+		try:
+			try:
+				signal.alarm(self._connect_timeout)
+				f = url_open(test_url)
+				early_out = True
+			finally:
+				signal.alarm(0)
+		except HTTPError as e:
+			self.output.write('deeptime(): connection to host %s\n'
+				'            returned HTTPError: %s for ip %s\n'
+				'            Switching back to original url\n'
+				% (url_parts.hostname, e, ip), 2)
+			if len(ips) == 1:
+				test_url = url_unparse(url_parts)
+				return self._test_connection(test_url, url_parts, ip, [])
+		except EnvironmentError as e:
+			self.output.write('deeptime(): connection to host %s '
+				'failed for ip %s:\n            %s\n'
+				% (url_parts.hostname, ip, e), 2)
+		except TimeoutException:
+			self.output.write(('deeptime(): connection to host %s '
+				'timed out for ip %s\n') % (url_parts.hostname, ip), 2)
+		return f, test_url, early_out
+
+
 	def _list_add(self, time_host, maxtime, host_dict, maxlen):
 		"""
 		Takes argumets ((time, host), maxtime, host_dict, maxlen)
 		Adds a new time:host pair to the dictionary of top hosts.
 		If the dictionary is full, the slowest host is removed to make space.
-		Returns the new maxtime, be it the specified timeout, or the slowest host.
+		Returns the new maxtime, be it the specified timeout,
+		or the slowest host.
 		"""
 		if len(host_dict) < maxlen:	#still have room, and host is fast. add it.
 
-			self.output.write('_list_add(): added host %s. with a time of %s\n' %
-					(time_host[1], time_host[0]), 2)
+			self.output.write('_list_add(): added host %s. with a time of %s\n'
+				% (time_host[1], time_host[0]), 2)
 
 			host_dict.update(dict([time_host]))
 			times = list(host_dict.keys())
 			times.sort()
 
 		else: #We need to make room in the dict before we add. Kill the slowest.
-			self.output.write('_list_add(): Adding host %s with a time of %s\n' %
-					(time_host[1], time_host[0]), 2)
+			self.output.write('_list_add(): Adding host %s with a time of %s\n'
+				% (time_host[1], time_host[0]), 2)
 			times = list(host_dict.keys())
 			times.sort()
-			self.output.write('_list_add(): removing %s\n' % host_dict[times[-1]],
-					2)
+			self.output.write('_list_add(): removing %s\n'
+				% host_dict[times[-1]], 2)
 			del host_dict[times[-1]]
 			host_dict.update(dict([time_host]))
 			#done adding. now return the appropriate time
@@ -476,8 +473,8 @@ class Deep(object):
 					' reusing timeout of %s sec.\n' % maxtime, 2)
 			retval = maxtime
 		else:
-			self.output.write('_list_add(): host_dict is full. Selecting the best'
-			' timeout\n', 2)
+			self.output.write('_list_add(): host_dict is full. '
+				'Selecting the best timeout\n', 2)
 			if times[-1] < maxtime:
 				retval = times[-1]
 			else:
@@ -497,7 +494,8 @@ class Interactive(object):
 		self.urls = []
 
 		self.interactive(hosts, options)
-		self.output.write('Interactive.interactive(): self.urls = %s\n' % self.urls, 2)
+		self.output.write('Interactive.interactive(): self.urls = %s\n'
+			% self.urls, 2)
 
 		if not self.urls or len(self.urls[0]) == 0:
 			sys.exit(1)
@@ -520,7 +518,8 @@ class Interactive(object):
 
 			dialog.extend(['20', '110', '14'])
 
-		for (url, args) in sorted(hosts, key = lambda x: (x[1]['country'].lower(), x[1]['name'].lower()) ):
+		for (url, args) in sorted(hosts, key = lambda x:
+				(x[1]['country'].lower(), x[1]['name'].lower()) ):
 			marker = ""
 			if options.rsync and not url.endswith("/gentoo-portage"):
 				url+="/gentoo-portage"
@@ -529,7 +528,8 @@ class Interactive(object):
 			if options.ipv6 and ( args['ipv6'] == 'n' ): continue
 			if options.ipv4 and ( args['ipv4'] == 'n' ): continue
 
-			#dialog.append('"%s" "%s%s: %s" "OFF"' % ( url, marker, args['country'], args['name']))
+			#dialog.append('"%s" "%s%s: %s" "OFF"'
+				#% ( url, marker, args['country'], args['name']))
 			dialog.extend(["%s" %url,
 				"%s%s: %s" %(marker, args['country'], args['name']),
 				 "OFF"])
@@ -543,7 +543,8 @@ class Interactive(object):
 
 		if self.urls:
 			if hasattr(self.urls[0], 'decode'):
-				self.urls = decode_selection([x.decode('utf-8').rstrip() for x in self.urls])
+				self.urls = decode_selection(
+					[x.decode('utf-8').rstrip() for x in self.urls])
 			else:
 				self.urls = decode_selection([x.rstrip() for x in self.urls])
 
